use std::{ffi::OsStr, io::ErrorKind, path::PathBuf};

use tokio::fs;
use tracing::{error, trace};
use url::Url;

pub fn as_path(url: &Url) -> PathBuf {
    // extract data from url to save it accurately
    let url_path = PathBuf::from("./downloaded/".to_string() + url.domain().unwrap_or("UnknownDomain") + url.path());

    // if it's a file
    let (basepath, filename) = if url_path.extension().filter(valid_file_extension).is_some() {
        // get everything up till the file
        let basepath = url_path.ancestors().skip(1).take(1).collect::<PathBuf>();
        // get the file name
        let filename = url_path.file_name().expect("This should exist").to_string_lossy();
        trace!("Save path: {:?} and base path: {:?}", &url_path, &basepath);
        (basepath, filename.to_string())
    } else {
        (url_path.clone(), "index.html".into())
    };

    let mut path = PathBuf::new();
    path = path.join(basepath);
    path = path.join(filename);

    path
}

pub async fn init(filename: &PathBuf) -> Option<fs::File> {
    let file = async || tokio::fs::OpenOptions::new()
        .append(true)
        .create(true)
        .open(&filename).await;

    match file().await {
        Ok(ok) => Some(ok),
        Err(err) => {
            // the file/folder isn't found
            if err.kind() == ErrorKind::NotFound {
                if let Some(parent ) = &filename.parent() {
                    // create the folders
                    if let Err(err) = fs::create_dir_all(&parent).await {
                        error!("Dir creation: {err} {:?}", filename);
                        eprintln!("{}", err)
                    } else if let Ok(ok) = file().await {
                        return Some(ok);
                    }
                } else {
                    error!("Couldn't get file's parents: {:?}", &filename);
                }
            } else {
                error!("File creation: {err} {:?}", filename);
            }
            // we don't care about other errors, we can't/shouldn't fix them
            None
        }
    }
}

fn valid_file_extension(take: &&OsStr) -> bool {
    let los = take.to_string_lossy();
    let all = los.split('.');
    match all.last() {
        Some(s) => {
            // FIXME it's worth noting that the dumb tlds like .zip are in here,
            // which could cause problems
            let all_domains = include_str!("tlds-alpha-by-domain.txt");

            // check if it is a domain
            match all_domains.lines().map(str::to_lowercase).find(|x| x==s.to_lowercase().as_str()) {
                Some(_) => false,
                None => true
            }
        },
        None => false,
    }
}
